# Record process driver
#
# Copyright 2008—2011, 2014 Mark Edgington <edgimar@gmail.com>
# Copyright 2016, 2018—2022 Andrej Shadura <andrew@shadura.me>
#
# This software may be used and distributed according to the terms of
# the GNU General Public License, incorporated herein by reference.
#
# Much of this extension is based on Bryan O'Sullivan's record extension.
#
# SPDX-License-Identifier: GPL-2.0-or-later

import errno
import io
import os
import subprocess
import tempfile
from gettext import gettext as _
from typing import IO, cast

from .chunk_selector import chunkselector
from .crpatch import Header, filterpatch, parsepatch
from .util import Abort, closefds, copyfile, system


def dorecord(ui, repo, *pats, **opts):
    """Drive the record process

    Interactively filter local changes, and accordingly prepare working
    directory into a state, where the job can be delegated to
    non-interactive commit command such as 'commit'.

    After the actual job is done by non-interactive command, working dir
    state is restored to original.

    In the end we'll record interesting changes, and everything else will be
    left in place, so the user can continue his work.

    We pass additional configuration options to Git to make the diff output
    consistent:
     - core.quotePath:
       Limit symbols requiring escaping to double-quotes, backslash and
       control characters.
     - diff.mnemonicPrefix:
       If set, git diff uses a prefix pair that is different from the standard
       "a/" and "b/" depending on what is being compared. Our parser only
       supports "a/" and "b/".
    """
    git_args = [
        "git",
        "-c",
        "core.quotePath=false",
        "-c",
        "diff.mnemonicPrefix=false",
        "diff",
        "--binary",
    ]
    git_base = []

    if opts['cached']:
        git_args.append("--cached")

    if not opts['index'] and repo.head():
        git_base.append("HEAD")

    p = subprocess.Popen(
        git_args + git_base, stdout=subprocess.PIPE, close_fds=closefds,
    )
    fp = cast(IO[bytes], p.stdout)

    # 0. parse patch
    fromfiles = set()
    tofiles = set()

    chunks = parsepatch(fp)
    for c in chunks:
        if isinstance(c, Header):
            fromfile, tofile = c.files()
            if fromfile is not None:
                fromfiles.add(os.fsdecode(fromfile))
            if tofile is not None:
                tofiles.add(os.fsdecode(tofile))

    added = tofiles - fromfiles
    removed = fromfiles - tofiles
    modified = tofiles - added - removed
    changes = [modified, added, removed]

    # 1. filter patch, so we have intending-to apply subset of it
    chunks = filterpatch(opts, chunks, chunkselector, ui)
    p.wait()
    del fp

    contenders = set()
    for h in chunks:
        fromfile, tofile = h.files()
        if fromfile is not None:
            contenders.add(os.fsdecode(fromfile))
        if tofile is not None:
            contenders.add(os.fsdecode(tofile))

    changed = changes[0] | changes[1] | changes[2]
    newfiles: list = [f for f in changed if f in contenders]

    if not newfiles:
        ui.status(_('no changes to record'))
        return 0

    # 2. backup changed files, so we can restore them in the end
    backups = {}
    newly_added_backups = {}
    backupdir = repo.controldir / 'record-backups'
    try:
        os.mkdir(backupdir)
    except OSError as err:
        if err.errno != errno.EEXIST:
            raise
    index_backup = None
    try:
        index_backup = repo.open_index()
        index_backup.backup_tree()

        # backup continues
        for f in newfiles:
            if f not in (modified | added):
                continue
            prefix = os.fsdecode(f).replace('/', '_') + '.'
            fd, tmpname = tempfile.mkstemp(
                prefix=prefix,
                dir=backupdir,
            )
            os.close(fd)
            ui.debug(f'backup {f!r} as {tmpname!r}')
            pathname = repo.path / f
            if os.path.isfile(pathname):
                copyfile(pathname, tmpname)
            if f in modified:
                backups[f] = tmpname
            elif f in added:
                newly_added_backups[f] = tmpname

        fp = io.BytesIO()
        all_backups = {}
        all_backups.update(backups)
        all_backups.update(newly_added_backups)
        for c in chunks:
            c.write(fp)
        dopatch = fp.tell()
        fp.seek(0)

        # 2.5 optionally review / modify patch in text editor
        if opts['crecord_reviewpatch']:
            patchtext = fp.read()
            reviewedpatch = ui.edit(patchtext, "")
            fp.truncate(0)
            fp.write(reviewedpatch)
            fp.seek(0)

        # 3a. apply filtered patch to clean repo  (clean)
        if backups or any(f in contenders for f in removed):
            system(
                ["git", "checkout", "-f"]
                + git_base
                + ["--"]
                + [f for f in newfiles if f not in added],
                onerr=Abort,
                errprefix=_("checkout failed"),
            )
        # remove newly added files from 'clean' repo (so patch can apply)
        for f in newly_added_backups:
            pathname = repo.path / f
            pathname.unlink(missing_ok=True)

        # 3b. (apply)
        if dopatch:
            try:
                ui.debug('applying patch')
                ui.debug(fp.getvalue().decode("UTF-8", "hexreplace"))
                p = subprocess.Popen(
                    ["git", "apply", "--whitespace=nowarn"],
                    stdin=subprocess.PIPE,
                    close_fds=closefds,
                )
                p.stdin.write(fp.getvalue())
                p.stdin.close()
                p.wait()
            except Exception as err:  # noqa: B902
                s = str(err)
                if s:
                    raise Abort(s)
                else:
                    raise Abort(_('patch failed to apply'))
        del fp

        # 4. We prepared working directory according to filtered patch.
        #    Now is the time to delegate the job to commit/qrefresh or the like!

        # it is important to first chdir to repo root -- we'll call a
        # highlevel command with list of pathnames relative to repo root
        newfiles = [repo.path / n for n in newfiles]
        if opts['operation'] == 'crecord':
            ui.commit(*newfiles, **opts)
        else:
            ui.stage(*newfiles, **opts)
        ui.debug('previous staging contents backed up as tree %r' % index_backup.indextree)
        index_backup = None

        return 0
    finally:
        # 5. finally restore backed-up files
        try:
            for realname, tmpname in backups.items():
                ui.debug(f'restoring {tmpname!r} to {realname!r}')
                copyfile(tmpname, os.path.join(repo.path, realname))
                os.unlink(tmpname)
            for realname, tmpname in newly_added_backups.items():
                ui.debug(f'restoring {tmpname!r} to {realname!r}')
                copyfile(tmpname, os.path.join(repo.path, realname))
                os.unlink(tmpname)
            os.rmdir(backupdir)
            if index_backup:
                index_backup.write()
        except (OSError, NameError):
            pass



from __future__ import annotations

import io
from collections.abc import Sequence
from functools import partial
from textwrap import dedent

import pytest

from git_crecord.crpatch import filterpatch, parsepatch


def line_selector(selections: Sequence[bool], opts, headers, ui):
    for i, hunkline in enumerate(headers[0].hunks[0].changedlines):
        hunkline.applied = selections[i]


@pytest.mark.parametrize(
    ("selections", "expected"),
    [
        pytest.param(
            [True, False, True, False],
            '''
            @@ -1,3 +1,3 @@
             RUN apt-get update
            - && apt-get install -y supervisor python3.8
            + && apt-get install -y supervisor python3.9
                 git python3-pip ssl-cert
            ''',
            id="symmetric",
        ),
        pytest.param(
            [True, True, True, False],
            '''
            @@ -1,3 +1,2 @@
             RUN apt-get update
            - && apt-get install -y supervisor python3.8
            + && apt-get install -y supervisor python3.9
            -    git python3-pip ssl-cert
            ''',
            id="extra deletion",
        ),
        pytest.param(
            [True, False, True, True],
            '''
            @@ -1,3 +1,4 @@
             RUN apt-get update
            - && apt-get install -y supervisor python3.8
            + && apt-get install -y supervisor python3.9
                 git python3-pip ssl-cert
            +    git python3-pip ssl-cert time
            ''',
            id="extra addition",
        ),
    ],
)
def test_hunk_splitting(selections: Sequence[bool], expected: str):
    diff = dedent(
        '''
        diff --git a/Dockerfile b/Dockerfile
        index 00083f466d4f..400252a22712 100644
        --- a/Dockerfile
        +++ b/Dockerfile
        @@ -1,3 +1,3 @@
         RUN apt-get update
        - && apt-get install -y supervisor python3.8
        -    git python3-pip ssl-cert
        + && apt-get install -y supervisor python3.9
        +    git python3-pip ssl-cert time
        ''',
    ).lstrip('\n').encode()
    patch = parsepatch(io.BytesIO(diff))
    applied = filterpatch(None, patch, partial(line_selector, selections), None)

    assert str(applied.hunks[0]) == dedent(expected).lstrip('\n')



# Git wrapper for repo/tree/index manipulation
#
# Copyright 2016, 2018—2022 Andrej Shadura <andrew@shadura.me>
#
# This software may be used and distributed according to the terms of
# the GNU General Public License, incorporated herein by reference.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from __future__ import annotations

import os
import sys
from pathlib import Path

from . import util

INDEX_FILENAME = "index"

ObjectHash = str


class GitTree:
    def __init__(self, tree):
        self._tree = tree

    def __repr__(self):
        return f"{self.__class__.__name__}({self._tree!r})"

    def read(self):
        util.system(['git', 'read-tree', '--reset', self._tree], onerr=RuntimeError)


class GitIndex:
    def __init__(self, filename):
        self._filename = filename
        self.indextree = None

    def __repr__(self):
        return f"{self.__class__.__name__}({self._filename!r}, {self.indextree!r})"

    def commit(self) -> ObjectHash:
        return util.systemcall(
            ['git', 'write-tree'],
            onerr=RuntimeError,
            encoding="ascii",
        ).rstrip('\n')

    def write(self):
        GitTree(self.indextree).read()

    def backup_tree(self) -> ObjectHash:
        try:
            self.indextree = self.commit()
        except RuntimeError as inst:
            raise util.Abort('failed to read the index: %s' % inst)
        return self.indextree


class GitRepo:
    def __init__(self, path: os.PathLike | str | None):
        try:
            self.path = Path(
                util.systemcall(
                    ['git', 'rev-parse', '--show-toplevel'],
                    dir=path,
                    encoding="fs",
                    onerr=util.Abort,
                ).rstrip('\n'),
            )
            self._controldir = Path(
                util.systemcall(
                    ['git', 'rev-parse', '--git-dir'],
                    dir=path,
                    encoding="fs",
                ).rstrip('\n'),
            )
            if not self._controldir.is_dir():
                raise util.Abort
        except util.Abort:
            sys.exit(1)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.path})"

    @property
    def controldir(self) -> Path:
        return self._controldir.resolve()

    @property
    def index_path(self) -> Path:
        return self.controldir / INDEX_FILENAME

    def open_index(self) -> GitIndex:
        return GitIndex(self.index_path)

    def head(self) -> ObjectHash:
        return util.systemcall(
            ['git', 'rev-parse', '--verify', '-q', 'HEAD'],
            encoding="ascii",
        ).rstrip('\n')



# Git diff parser and related structures
#
# Copyright 2008—2011, 2014 Mark Edgington <edgimar@gmail.com>
# Copyright 2016, 2018—2022 Andrej Shadura <andrew@shadura.me>
#
# This software may be used and distributed according to the terms of
# the GNU General Public License, incorporated herein by reference.
#
# This code is based on the patch parser of Mercurial.
#
# SPDX-License-Identifier: GPL-2.0-or-later

import io
import re
from codecs import register_error
from gettext import gettext as _
from typing import IO, Optional, Union
from collections.abc import Iterator, Sequence

from .util import unwrap_filename

lines_re = re.compile(b'@@ -(\\d+)(?:,(\\d+))? \\+(\\d+)(?:,(\\d+))? @@\\s*(.*)')


class PatchError(Exception):
    pass


def hexreplace(err: UnicodeError) -> tuple[str, int]:
    if not isinstance(err, UnicodeDecodeError):
        raise NotImplementedError("only decoding is supported")
    return "".join(
        "<%X>" % x for x in err.object[err.start:err.end]
    ), err.end


register_error("hexreplace", hexreplace)


class LineReader:
    # simple class to allow pushing lines back into the input stream
    def __init__(self, fp: IO[bytes]):
        self.fp = fp
        self.buf: list[bytes] = []

    def push(self, line: bytes) -> None:
        if line is not None:
            self.buf.append(line)

    def readline(self) -> bytes:
        if self.buf:
            line = self.buf[0]
            del self.buf[0]
            return line
        return self.fp.readline()

    def __iter__(self) -> Iterator[bytes]:
        return iter(self.readline, b'')


def scanpatch(fp: IO[bytes]):
    r"""Read a patch and yield the following events:

    - ('file',    [header_lines + fromfile + tofile])
    - ('context', [context_lines])
    - ('hunk',    [hunk_lines])
    - ('range',   (-start,len, +start,len, diffp))

    >>> rawpatch = b'''diff --git a/folder1/g b/folder1/g
    ... --- a/folder1/g
    ... +++ b/folder1/g
    ... @@ -1,8 +1,10 @@ some context
    ...  1
    ...  2
    ... -3
    ...  4
    ...  5
    ...  6
    ... +6.1
    ... +6.2
    ...  7
    ...  8
    ... +9'''
    >>> fp = io.BytesIO(rawpatch)
    >>> list(scanpatch(fp))
    [('file',
        [b'diff --git a/folder1/g b/folder1/g\n',
         b'--- a/folder1/g\n',
         b'+++ b/folder1/g\n']),
     ('range',
        (b'1', b'8', b'1', b'10', b'some context')),
     ('context',
        [b' 1\n', b' 2\n']),
     ('hunk',
        [b'-3\n']),
     ('context',
        [b' 4\n', b' 5\n', b' 6\n']),
     ('hunk',
        [b'+6.1\n', b'+6.2\n']),
     ('context',
        [b' 7\n', b' 8\n']),
     ('hunk',
        [b'+9'])]
    """
    lr = LineReader(fp)

    def scanwhile(first: bytes, p) -> list[bytes]:
        """scan lr while predicate holds"""
        lines = [first]
        for line in iter(lr.readline, b''):
            if p(line):
                lines.append(line)
            else:
                lr.push(line)
                break
        return lines

    for line in iter(lr.readline, b''):
        if line.startswith(b'diff --git a/') or line.startswith(b'diff --git "a/'):
            def notheader(line: bytes) -> bool:
                s = line.split(None, 1)
                return not s or s[0] not in (b'---', b'diff')

            header = scanwhile(line, notheader)
            fromfile = lr.readline()
            if fromfile.startswith(b'---'):
                tofile = lr.readline()
                header += [fromfile, tofile]
            else:
                lr.push(fromfile)
            yield 'file', header
        elif line.startswith(b' '):
            yield 'context', scanwhile(line, lambda l: l[0] in b' ')
        elif line[0] in b'-+':
            yield 'hunk', scanwhile(line, lambda l: l[0] in b'-+\\')
        elif line.startswith(b'* Unmerged path '):
            continue
        else:
            m = lines_re.match(line)
            if m:
                yield 'range', m.groups()
            else:
                raise PatchError('unknown patch content: %r' % line)


class PatchNode:
    """Abstract Class for Patch Graph Nodes
    (i.e. PatchRoot, header, hunk, HunkLine)
    """

    folded: bool
    applied: bool
    # a patch this node belongs to
    patch: 'PatchRoot'

    def firstchild(self):
        raise NotImplementedError("method must be implemented by subclass")

    def lastchild(self):
        raise NotImplementedError("method must be implemented by subclass")

    def allchildren(self) -> Sequence['PatchNode']:
        """Return a list of all direct children of this node"""
        raise NotImplementedError("method must be implemented by subclass")

    def nextsibling(self) -> Optional['PatchNode']:
        """
        Return the closest next item of the same type where there are no items
        of different types between the current item and this closest item.
        If no such item exists, return None.
        """
        raise NotImplementedError("method must be implemented by subclass")

    def prevsibling(self) -> Optional['PatchNode']:
        """
        Return the closest previous item of the same type where there are no
        items of different types between the current item and this closest item.
        If no such item exists, return None.
        """
        raise NotImplementedError("method must be implemented by subclass")

    def parentitem(self) -> Optional['PatchNode']:
        """Return the parent to the current item"""
        raise NotImplementedError("method must be implemented by subclass")

    def nextitem(self, skipfolded=True) -> Optional['PatchNode']:
        """
        Try to return the next item closest to this item, regardless of item's
        type (header, hunk, or hunkline).

        If skipfolded == True, and the current item is folded, then the child
        items that are hidden due to folding will be skipped when determining
        the next item.

        If it is not possible to get the next item, return None.
        """
        try:
            itemfolded = self.folded
        except AttributeError:
            itemfolded = False
        if skipfolded and itemfolded:
            nextitem = self.nextsibling()
            if not nextitem:
                parent = self.parentitem()
                if parent:
                    nextitem = parent.nextsibling()
                else:
                    nextitem = None
            return nextitem
        else:
            # try child
            item = self.firstchild()
            if item:
                return item

            # else try next sibling
            item = self.nextsibling()
            if item:
                return item

            parent = self.parentitem()
            if parent:
                # else try parent's next sibling
                item = parent.nextsibling()
                if item:
                    return item

                # else return grandparent's next sibling (or None)
                grandparent = parent.parentitem()
                if grandparent:
                    return grandparent.nextsibling()
                else:
                    return None

            else:  # parent and/or grandparent was None
                return None

    def previtem(self) -> Optional['PatchNode']:
        """
        Try to return the previous item closest to this item, regardless of
        item's type (header, hunk, or hunkline).

        If it is not possible to get the previous item, return None.
        """
        # try previous sibling's last child's last child,
        # else try previous sibling's last child, else try previous sibling
        prevsibling = self.prevsibling()
        if prevsibling is not None:
            prevsiblinglastchild = prevsibling.lastchild()
            if (prevsiblinglastchild is not None) and not prevsibling.folded:
                prevsiblinglclc = prevsiblinglastchild.lastchild()
                if (prevsiblinglclc is not None) and not prevsiblinglastchild.folded:
                    return prevsiblinglclc
                else:
                    return prevsiblinglastchild
            else:
                return prevsibling

        # try parent (or None)
        return self.parentitem()

    def write(self, fp: IO[bytes]) -> None:
        """Write the unified diff-formatter representation of the
        patch node into the binary stream"""
        raise NotImplementedError("method must be implemented by subclass")

    def __bytes__(self) -> bytes:
        """Return the unified diff-formatter representation of the
        patch node as bytes"""
        with io.BytesIO() as b:
            self.write(b)
            return b.getvalue()


class Header(PatchNode):
    """Patch header"""

    diff_re = re.compile(b'diff --git (?P<fromfile>(?P<aq>")?a/.*(?(aq)"|)) (?P<tofile>(?P<bq>")?b/.*(?(bq)"|))$')
    allhunks_re = re.compile(b'(?:GIT binary patch|new file|deleted file) ')
    pretty_re = re.compile(b'(?:new file|deleted file) ')
    special_re = re.compile(b'(?:GIT binary patch|new|deleted|copy|rename) ')

    def __init__(self, header):
        self.header = header
        self.hunks = []
        # flag to indicate whether to apply this chunk
        self.applied = True
        # flag which only affects the status display indicating if a node's
        # children are partially applied (i.e. some applied, some not).
        self.partial = False

        # flag to indicate whether to display as folded/unfolded to user
        self.folded = True

        # flag is False if this header was ever unfolded from initial state
        self.neverunfolded = True

        # one-letter file status
        self._changetype = None

    def binary(self):
        """Return True if the file represented by the header is a binary file."""
        return any(h.startswith(b'GIT binary patch') for h in self.header)

    def pretty(self, fp: IO[str]):
        """Pretty-print the header into a stream"""
        for h in self.header:
            if h.startswith(b'GIT binary patch'):
                fp.write(_('this modifies a binary file (all or nothing)\n'))
                break
            if self.pretty_re.match(h):
                fp.write(h.decode("UTF-8", errors="hexreplace"))
                if self.binary():
                    fp.write(_('this is a binary file\n'))
                break
            if h.startswith(b'---'):
                fp.write(
                    _('%d hunks, %d lines changed\n') %
                    (
                        len(self.hunks),
                        sum(max(h.added, h.removed) for h in self.hunks),
                    ),
                )
                break
            fp.write(h.decode("UTF-8", errors="hexreplace"))

    def prettystr(self) -> str:
        return str(self)

    def __str__(self) -> str:
        with io.StringIO() as s:
            self.pretty(s)
            return s.getvalue()

    def write(self, fp: IO[bytes]) -> None:
        fp.write(b''.join(self.header))

    def allhunks(self) -> bool:
        """
        Return True if the file which the header represents was changed
        completely (i.e.  there is no possibility of applying a hunk of changes
        smaller than the size of the entire file.)  Otherwise, return False
        """
        return any(self.allhunks_re.match(h) for h in self.header)

    def files(self):
        fromfile, tofile = self.diff_re.match(self.header[0]).group('fromfile', 'tofile')
        fromfile = unwrap_filename(fromfile).removeprefix(b'a/')
        tofile = unwrap_filename(tofile).removeprefix(b'b/')
        if self.changetype == 'D':
            tofile = None
        elif self.changetype == 'A':
            fromfile = None
        return [fromfile, tofile]

    def filename(self) -> str:
        files = self.files()
        return (files[1] or files[0]).decode("UTF-8", errors="hexreplace")

    def __repr__(self) -> str:
        return '<header %s>' % (
            ' '.join(
                repr(x) for x in self.files()
            )
        )

    def special(self) -> bool:
        return any(self.special_re.match(h) for h in self.header)

    @property
    def changetype(self) -> str:
        if self._changetype is None:
            self._changetype = "M"
            for h in self.header:
                if h.startswith(b'new file'):
                    self._changetype = "A"
                elif h.startswith(b'deleted file'):
                    self._changetype = "D"
                elif h.startswith(b'copy from'):
                    self._changetype = "C"
                elif h.startswith(b'rename from'):
                    self._changetype = "R"

        return self._changetype

    def nextsibling(self) -> Optional['Header']:
        numheadersinpatch = len(self.patch)
        indexofthisheader = self.patch.index(self)

        if indexofthisheader < numheadersinpatch - 1:
            nextheader = self.patch[indexofthisheader + 1]
            return nextheader
        else:
            return None

    def prevsibling(self) -> Optional['Header']:
        indexofthisheader = self.patch.index(self)
        if indexofthisheader > 0:
            previousheader = self.patch[indexofthisheader - 1]
            return previousheader
        else:
            return None

    def parentitem(self) -> None:
        """
        There is no 'real' parent item of a header that can be selected,
        so return None.
        """
        return None

    def firstchild(self):
        """Return the first child of this item, if one exists.  Otherwise, None."""
        if len(self.hunks) > 0:
            return self.hunks[0]
        else:
            return None

    def lastchild(self):
        """Return the last child of this item, if one exists.  Otherwise, None."""
        if len(self.hunks) > 0:
            return self.hunks[-1]
        else:
            return None

    def allchildren(self) -> Sequence['Hunk']:
        """Return a list of all direct children of this node"""
        return self.hunks


class HunkLine(PatchNode):
    """Represents a changed line in a hunk"""

    DELETE = b'-'
    INSERT = b'+'
    CONTEXT = b' '
    HUNK = b'@'
    NOEOL = b'\\'

    linetext: bytes
    hunk: 'Hunk'
    offset: int

    def __init__(self, linetext: bytes, hunk: 'Hunk'):
        self.linetext = linetext
        self.applied = True
        # the parent hunk to which this line belongs
        self.hunk = hunk
        # folding lines currently is not used/needed, but this flag is needed
        # in the prevItem method.
        self.folded = False
        # at this moment, the offset of the line in the hunk isn’t known yet
        self.offset = 0

    def __bytes__(self):
        if self.applied or self.diffop == HunkLine.NOEOL:
            return self.linetext
        else:
            return b' ' + self.linetext[1:]

    @property
    def diffop(self):
        return self.linetext[0:1]

    def __repr__(self):
        return "<hunkline/%c/%d %s %s>" % (
            self.linetext[0],
            self.offset,
            '[x]' if self.applied else '[ ]',
            self.linetext[1:10],
        )

    def __str__(self) -> str:
        return self.prettystr()

    def prettystr(self) -> str:
        return self.linetext.decode("UTF-8", errors="hexreplace")

    def nextsibling(self):
        numlinesinhunk = len(self.hunk.changedlines)
        indexofthisline = self.hunk.changedlines.index(self)

        if indexofthisline < numlinesinhunk - 1:
            nextline = self.hunk.changedlines[indexofthisline + 1]
            return nextline
        else:
            return None

    def prevsibling(self) -> Optional['HunkLine']:
        """Return the previous line in the hunk"""
        indexofthisline = self.hunk.changedlines.index(self)
        if indexofthisline > 0:
            previousline = self.hunk.changedlines[indexofthisline - 1]
            return previousline
        else:
            return None

    def parentitem(self) -> 'Hunk':
        """Return the parent to the current item"""
        return self.hunk

    def firstchild(self):
        """Return the first child of this item, if one exists.  Otherwise, None."""
        # hunk-lines don't have children
        return None

    def lastchild(self):
        """Return the last child of this item, if one exists.  Otherwise, None."""
        # hunk-lines don't have children
        return None


class Hunk(PatchNode):
    """ui patch hunk, wraps a hunk and keeps track of ui behavior """

    maxcontext = 3
    header: Header
    fromline: int
    toline: int
    proc: bytes
    after: Sequence[bytes]
    before: Sequence[bytes]
    changedlines: Sequence[HunkLine]

    def __init__(
        self,
        header: Header,
        fromline: int,
        toline: int,
        proc: bytes,
        before: Sequence[bytes],
        hunklines: Sequence[bytes],
        after: Sequence[bytes],
    ):
        def trimcontext(number, lines):
            delta = len(lines) - self.maxcontext
            # TODO: why is this disabled?
            if False and delta > 0:
                return number + delta, lines[:self.maxcontext]
            return number, lines

        self.header = header
        self.fromline, self.before = trimcontext(fromline, before)
        self.toline, self.after = trimcontext(toline, after)
        self.proc = proc
        self.changedlines = [HunkLine(line, self) for line in hunklines]
        self.added, self.removed = self.countchanges()
        self.countoffsets()
        # used at end for detecting how many removed lines were un-applied
        self.originalremoved = self.removed

        # flag to indicate whether to display as folded/unfolded to user
        self.folded = True
        # flag to indicate whether to apply this chunk
        self.applied = True
        # flag which only affects the status display indicating if a node's
        # children are partially applied (i.e. some applied, some not).
        self.partial = False

    def nextsibling(self) -> Optional['Hunk']:
        """Return the next hunk in the group."""
        numhunksinheader = len(self.header.hunks)
        indexofthishunk = self.header.hunks.index(self)

        if indexofthishunk < numhunksinheader - 1:
            nexthunk = self.header.hunks[indexofthishunk + 1]
            return nexthunk
        else:
            return None

    def prevsibling(self) -> Optional['Hunk']:
        """Return the previous hunk in the group."""
        indexofthishunk = self.header.hunks.index(self)
        if indexofthishunk > 0:
            previoushunk = self.header.hunks[indexofthishunk - 1]
            return previoushunk
        else:
            return None

    def parentitem(self) -> Header:
        """Return the header for this hunk"""
        return self.header

    def firstchild(self) -> Optional[HunkLine]:
        """Return the first hunk line of this hunk, if one exists.  Otherwise, None."""
        if len(self.changedlines) > 0:
            return self.changedlines[0]
        else:
            return None

    def lastchild(self) -> Optional[HunkLine]:
        """Return the last child of this item, if one exists.  Otherwise, None."""
        if len(self.changedlines) > 0:
            return self.changedlines[-1]
        else:
            return None

    def allchildren(self) -> Sequence[PatchNode]:
        """Return a list of all direct children of this node"""
        return self.changedlines

    def countchanges(self) -> tuple[int, int]:
        """changedlines -> (n+,n-)"""
        add = len(
            [line for line in self.changedlines if line.applied and line.diffop == HunkLine.INSERT],
        )
        rem = len(
            [line for line in self.changedlines if line.applied and line.diffop == HunkLine.DELETE],
        )
        return add, rem

    def countoffsets(self):
        fromline = 0
        toline = 0
        deletes = False
        for line in self.changedlines:
            if line.diffop == HunkLine.INSERT:
                fromline += 1
                line.offset = fromline
            if line.diffop == HunkLine.DELETE:
                deletes = True
                toline += 1
                line.offset = toline
            if line.diffop == HunkLine.NOEOL:
                if not deletes:
                    line.offset = fromtoline
                else:
                    line.offset = toline

    def getfromtoline(self):
        """Calculate the number of removed lines converted to context lines"""
        removedconvertedtocontext = self.originalremoved - self.removed

        contextlen = (
            len(self.before) + len(self.after) +
            removedconvertedtocontext
        )
        if self.after and self.after[-1] == b'\\ No newline at end of file\n':
            contextlen -= 1
        fromlen = contextlen + self.removed
        tolen = contextlen + self.added

        # Diffutils manual, section "2.2.2.2 Detailed Description of Unified
        # Format": "An empty hunk is considered to end at the line that
        # precedes the hunk."
        #
        # So, if either of hunks is empty, decrease its line start. --immerrr
        # But only do this if fromline > 0, to avoid having, e.g fromline=-1.
        fromline, toline = self.fromline, self.toline
        if fromlen == 0 and fromline > 0:
            fromline -= 1
        if tolen == 0 and toline > 0:
            toline -= 1

        fromtoline = b"@@ -%d,%d +%d,%d @@%b\n" % (
            fromline,
            fromlen,
            toline,
            tolen,
            self.proc and (b" " + self.proc),
        )

        return fromtoline

    def iterlines(self, all: bool = False):
        # updated self.added/removed, which are used by getfromtoline()
        self.added, self.removed = self.countchanges()
        yield HunkLine.HUNK, self.getfromtoline()
        for line in self.before:
            yield HunkLine.CONTEXT, line

        # Include these lines:
        # (1) all applied lines
        # (2) all unapplied removal lines (converted context lines)
        # Lines are sorted by their offset in the original hunk,
        # with removals coming first.
        for hunkline in sorted(
            self.changedlines,
            key=lambda line: (
                line.offset, line.diffop == HunkLine.INSERT,
            ),
        ):
            if hunkline.diffop == HunkLine.INSERT and not (hunkline.applied or all):
                continue

            yield hunkline.diffop, bytes(hunkline)

        for line in self.after:
            yield HunkLine.CONTEXT, line

    def write(self, fp: IO[bytes]) -> None:
        for _, line in self.iterlines():
            fp.write(line)

    def reversehunks(self) -> 'Hunk':
        r"""Make the hunk apply in the other direction.

        >>> header = Header([b'diff --git a/file b/file\n'])
        >>> print(Hunk(
        ...     header,
        ...     fromline=1,
        ...     toline=2,
        ...     proc=b'context',
        ...     before=[b' 1\n', b' 2\n'],
        ...     hunklines=[b'-3\n'],
        ...     after=[b' 4\n', b' 5\n'],
        ... ).reversehunks().prettystr())
        @@ -1,4 +2,5 @@ context
         1
         2
        +3
         4
         5
        """
        m = {b"+": b"-", b"-": b"+", b"\\": b"\\"}
        hunklines = [
            b"%s%s" % (m[line.linetext[0:1]], line.linetext[1:])
            for line in self.changedlines
            if line.applied
        ]
        return Hunk(
            self.header,
            self.fromline,
            self.toline,
            self.proc,
            self.before,
            hunklines,
            self.after,
        )

    def files(self) -> list[Optional[bytes]]:
        return self.header.files()

    def filename(self) -> str:
        return self.header.filename()

    def __str__(self) -> str:
        return self.prettystr()

    def prettystr(self) -> str:
        x = io.BytesIO()
        self.write(x)
        return x.getvalue().decode("UTF-8", errors="hexreplace")

    def __repr__(self) -> str:
        return '<hunk %r@%d>' % (self.files()[1] or self.files()[0], self.fromline)


class PatchRoot(PatchNode, list):
    """List of header objects representing the patch."""

    def __init__(self, headerlist):
        super().__init__()
        self.extend(headerlist)
        # add parent patch object reference to each header
        for header in self:
            header.patch = self

    @property
    def headers(self) -> Sequence[Header]:
        return [c for c in self if isinstance(c, Header)]

    @property
    def hunks(self) -> Sequence[Hunk]:
        return [c for c in self if isinstance(c, Hunk)]


def parsepatch(fp: IO[bytes]) -> PatchRoot:
    r"""Parse a patch, returning a list of header and hunk objects.

    >>> rawpatch = b'''diff --git a/folder1/g b/folder1/g
    ... --- a/folder1/g
    ... +++ b/folder1/g
    ... @@ -1,8 +1,10 @@
    ...  1
    ...  2
    ... -3
    ...  4
    ...  5
    ...  6
    ... +6.1
    ... +6.2
    ...  7
    ...  8
    ... +9'''
    >>> fp = io.BytesIO(rawpatch)
    >>> headers = parsepatch(fp)

    Headers and hunks are interspersed in the list returned from
    the function:
    >>> headers
    [<header b'folder1/g' b'folder1/g'>,
     <hunk b'folder1/g'@1>,
     <hunk b'folder1/g'@7>,
     <hunk b'folder1/g'@9>]

    >>> headers[0].filename()
    'folder1/g'

    Each header also provides a list of hunks belonging to it:
    >>> headers[0].hunks
    [<hunk b'folder1/g'@1>,
     <hunk b'folder1/g'@7>,
     <hunk b'folder1/g'@9>]
    >>> out = io.BytesIO()
    >>> for header in headers:
    ...     header.write(out)
    >>> print(out.getvalue().decode("ascii"))
    diff --git a/folder1/g b/folder1/g
    --- a/folder1/g
    +++ b/folder1/g
    @@ -1,6 +1,5 @@
     1
     2
    -3
     4
     5
     6
    @@ -7,2 +6,4 @@
    +6.1
    +6.2
     7
     8
    @@ -8,0 +10,1 @@
    +9

    Parsing a patch changing the EOL at the EOF:
    >>> rawpatch = b'''diff --git a/test b/test
    ... --- a/test
    ... +++ b/test
    ... @@ -11,5 +11,5 @@
    ...  foo
    ...  bar
    ...  baz
    ... -qoox
    ... -qooox
    ... \\ No newline at end of file
    ... +quux
    ... +quuux'''
    >>> fp = io.BytesIO(rawpatch)
    >>> headers = parsepatch(fp)
    >>> print(headers[0].hunks[0])
    @@ -11,5 +11,5 @@
     foo
     bar
     baz
    -qoox
    +quux
    -qooox
    \ No newline at end of file
    +quuux

    It is possible to handle non-UTF-8 patches:
    >>> rawpatch = b'''diff --git a/test b/test
    ... --- /dev/null
    ... +++ b/test
    ... @@ -0,0 +1,2 @@
    ... +\xCD\xCE\xCD-\xD3\xD2\xD4-8 \xF2\xE5\xF1\xF2
    ... +test'''
    >>> fp = io.BytesIO(rawpatch)
    >>> headers = parsepatch(fp)
    >>> out = io.BytesIO()
    >>> for header in headers:
    ...     header.write(out)

    Non-UTF-8 characters survive the roundtrip:
    >>> print(out.getvalue().decode("cp1251"))
    diff --git a/test b/test
    --- /dev/null
    +++ b/test
    @@ -0,0 +1,2 @@
    +НОН-УТФ-8 тест
    +test

    When pretty-printing the hunk, they get replaced with their
    hexadecimal codes:
    >>> print(headers[0].hunks[0])
    @@ -0,0 +1,2 @@
    +<CD><CE><CD>-<D3><D2><D4>-8 <F2><E5><F1><F2>
    +test

    Quoted filenames in the diff headers are supported too:
    >>> rawpatch = b'''diff --git "a/test- \\\\\\321\\217\\321\\217" "b/test- \\\\\\321\\217\\321\\217"
    ... new file mode 100644
    ... index 000000000000..7f53c853ca78
    ... --- /dev/null
    ... +++ "b/test- \\\\\\321\\217\\321\\217"
    ... @@ -0,0 +1,2 @@
    ... +\xCD\xCE\xCD-\xD3\xD2\xD4-8 \xF2\xE5\xF1\xF2
    ... +test'''
    >>> fp = io.BytesIO(rawpatch)
    >>> patch = parsepatch(fp)
    >>> files = patch.headers[0].files()
    >>> files[0]
    >>> files[1].decode('UTF-8')
    'test- \\яя'
    """

    class Parser:
        """patch parsing state machine"""

        header: Header
        headers: Sequence[Union[Header, Hunk]]

        def __init__(self):
            self.fromline = 0
            self.toline = 0
            self.proc = b''
            self.context = []
            self.before = []
            self.hunk = []
            self.headers = []

        def addrange(self, limits):
            """Store range line info to associated instance variables."""
            fromstart, fromend, tostart, toend, proc = limits
            self.fromline = int(fromstart)
            self.toline = int(tostart)
            self.proc = proc

        def add_new_hunk(self):
            """
            Create a new complete hunk object, adding it to the latest header
            and to self.headers.

            Add all of the previously collected information about
            the hunk to the new hunk object.  This information includes
            header, from/to-lines, function (self.proc), preceding context
            lines, changed lines, as well as the current context lines (which
            follow the changed lines).

            The size of the from/to lines are updated to be correct for the
            next hunk we parse.

            """
            h = Hunk(
                self.header,
                self.fromline,
                self.toline,
                self.proc,
                self.before,
                self.hunk,
                self.context,
            )
            self.header.hunks.append(h)
            self.headers.append(h)
            self.fromline += len(self.before) + h.removed + len(self.context)
            self.toline += len(self.before) + h.added + len(self.context)
            self.before = []
            self.hunk = []
            self.context = []
            self.proc = b''

        def addcontext(self, context):
            """
            Set the value of self.context.

            Also, if an unprocessed set of changelines was previously
            encountered, this is the condition for creating a complete
            hunk object.  In this case, we create and add a new hunk object to
            the most recent header object, and to self.strem.

            """
            self.context = context
            # if there have been changed lines encountered that haven't yet
            # been add to a hunk.
            if self.hunk:
                self.add_new_hunk()

        def addhunk(self, hunk):
            """
            Store the changed lines in self.changedlines.

            Mark any context lines in the context-line buffer (self.context) as
            lines preceding the changed-lines (i.e. stored in self.before), and
            clear the context-line buffer.

            """
            self.hunk = hunk
            self.before = self.context
            self.context = []

        def newfile(self, header):
            """
            Create a header object containing the header lines, and the
            filename the header applies to.  Add the header to self.headers.

            """
            # if there are any lines in the unchanged-lines buffer, create a
            # new hunk using them, and add it to the last header.
            if self.hunk:
                self.add_new_hunk()

            # create a new header and add it to self.header
            h = Header(header)
            self.headers.append(h)
            self.header = h

        def finished(self):
            # if there are any lines in the unchanged-lines buffer, create a
            # new hunk using them, and add it to the last header.
            if self.hunk:
                self.add_new_hunk()

            return self.headers

        transitions = {
            'file': {
                'context': addcontext,
                'file': newfile,
                'hunk': addhunk,
                'range': addrange,
            },
            'context': {
                'file': newfile,
                'hunk': addhunk,
                'range': addrange,
            },
            'hunk': {
                'context': addcontext,
                'file': newfile,
                'range': addrange,
            },
            'range': {
                'context': addcontext,
                'hunk': addhunk,
            },
        }

    p = Parser()

    # run the state-machine
    state = 'context'
    for newstate, data in scanpatch(fp):
        try:
            p.transitions[state][newstate](p, data)
        except KeyError:
            raise PatchError(
                'unhandled transition: %s -> %s' %
                (state, newstate),
            )
        state = newstate
    return PatchRoot(p.finished())


def filterpatch(opts, patch: PatchRoot, chunkselector, ui):
    r"""Interactively filter patch chunks into applied-only chunks

    >>> rawpatch = b'''* Unmerged path .gitignore
    ... diff --git a/dir/file.c b/dir/file.c
    ... index e548702cb275..28208f7ff2ac 100644
    ... --- a/dir/file.c
    ... +++ b/dir/file.c
    ... @@ -1684,11 +1684,13 @@ int function()
    ...  1
    ...  2
    ...  3
    ... -4
    ... -5
    ... -6
    ... -7
    ...  8
    ... +9
    ... +10
    ... +11
    ... +12
    ... +13
    ... +14
    ...  15
    ...  16
    ...  17
    ... '''
    >>> from functools import partial
    >>> def hunk_selector(selections: Sequence[bool], opts, headers, ui):
    ...     for i, hunk in enumerate(headers[0].hunks):
    ...         hunk.applied = selections[i]
    ...
    >>> def line_selector(selections: Sequence[bool], opts, headers, ui):
    ...     for i, hunkline in enumerate(headers[0].hunks[0].changedlines):
    ...         hunkline.applied = selections[i]
    ...
    >>> patch = parsepatch(io.BytesIO(rawpatch))
    >>> patch
    [<header b'dir/file.c' b'dir/file.c'>,
     <hunk b'dir/file.c'@1684>,
     <hunk b'dir/file.c'@1692>]
    >>> selections = [False, True]
    >>> applied = filterpatch(None, patch, partial(hunk_selector, selections), None)
    >>> applied
    [<header b'dir/file.c' b'dir/file.c'>,
     <hunk b'dir/file.c'@1692>]
    >>> print(applied.hunks[0])
    @@ -1692,3 +1692,9 @@
    +9
    +10
    +11
    +12
    +13
    +14
     15
     16
     17
    """
    # if there are no changed files
    if len(patch) == 0:
        return []

    # let user choose headers/hunks/lines, and mark their applied flags accordingly
    chunkselector(opts, patch.headers, ui)

    applied_hunks = PatchRoot([])
    for header in patch.headers:
        if header.applied and (
            header.special()
            or header.binary()
            or len([h for h in header.hunks if h.applied]) > 0
        ):
            applied_hunks.append(header)
            fixoffset = 0
            for hunk in header.hunks:
                if hunk.applied:
                    applied_hunks.append(hunk)
                    # adjust the 'to'-line offset of the hunk to be correct
                    # after de-activating some other hunks for this file
                    if fixoffset:
                        # hunk = copy.copy(hunk) # necessary??
                        hunk.toline += fixoffset
                else:
                    fixoffset += hunk.removed - hunk.added

    return applied_hunks



# Unicode string width calculator
#
#  Copyright 2009, 2010 Matt Mackall <mpm@selenic.com>
#  Copyright 2010, 2011 FUJIWARA Katsunori <foozy@lares.dti.ne.jp>
#  Copyright 2011 Augie Fackler <durin42@gmail.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# SPDX-License-Identifier: GPL-2.0-or-later

import unicodedata

# How to treat ambiguous-width characters. Set to 'WFA' to treat as wide.
wide = "WF"


def ucolwidth(d: str) -> int:
    """Find the column width of a Unicode string for display"""
    eaw = getattr(unicodedata, 'east_asian_width', None)
    if eaw is not None:
        return sum(eaw(c) in wide and 2 or 1 for c in d)
    return len(d)



# git-crecord entry point and configuration helpers
#
# Copyright 2016, 2018—2022 Andrej Shadura <andrew@shadura.me>
#
# This software may be used and distributed according to the terms of
# the GNU General Public License, incorporated herein by reference.
#
# SPDX-License-Identifier: GPL-2.0-or-later

import argparse
import importlib.metadata
import os
import sys
import tempfile
from gettext import gettext as _
from pathlib import Path
from typing import Optional

from . import crecord_core
from .gitrepo import GitRepo
from .util import Abort, system, systemcall


class Config:
    def get(self, section, item, default=None) -> Optional[str]:
        try:
            return systemcall(
                ['git', 'config', '--get', f'{section}.{item}'],
                onerr=KeyError,
                encoding="UTF-8",
            ).rstrip('\n')
        except KeyError:
            return default

    def set(self, section, item, value, source=""):
        raise NotImplementedError


class Ui:
    def __init__(self, repo: GitRepo):
        self.repo = repo
        self.config = Config()
        self.debuglevel = 0

    def print_message(self, *msg, debuglevel: int, **opts):
        if self.debuglevel < debuglevel:
            return

        sys.stdout.flush()
        print(*msg, **opts, file=sys.stderr)
        sys.stderr.flush()

    def debug(self, *msg, **opts):
        self.print_message(*msg, debuglevel=2, **opts)

    def info(self, *msg, **opts):
        self.print_message(*msg, debuglevel=1, **opts)

    def warn(self, *msg, **opts):
        self.print_message(*msg, debuglevel=0, **opts)

    def status(self, *msg, **opts):
        print(*msg, **opts)

    def setdebuglevel(self, level):
        self.debuglevel = level

    @property
    def editor(self) -> str:
        return (
            os.environ.get("GIT_EDITOR")
            or self.config.get("core", "editor")
            or os.environ.get("VISUAL")
            or os.environ.get("EDITOR")
            or "sensible-editor"
        )

    def edit(self, text: bytes, user, extra=None, name=None) -> bytes:
        f = tempfile.NamedTemporaryFile(
            prefix='git-crecord-',
            suffix=".txt",
            mode="wb",
            delete=False,
        )
        try:
            f.write(text)
            f.close()

            editor = self.editor

            system(
                f'{editor} "{f.name}"', onerr=Abort, errprefix=_("edit failed"),
            )

            t = Path(f.name).read_bytes()

        finally:
            os.unlink(f.name)

        return t

    def stage(self, *files, **opts):
        to_add = [f for f in files if os.path.exists(f)]
        if to_add:
            system(
                ['git', 'add', '-f', '--'] + to_add,
                onerr=Abort,
                errprefix=_("add failed"),
            )

    def commit(self, *files, **opts):
        msgfile = self.repo.controldir / "CRECORD_COMMITMSG"
        try:
            args = []

            if opts["message"]:
                msgfile.write_text(opts["message"])

            if opts["cleanup"] is None:
                opts["cleanup"] = "strip"

            for k, v in opts.items():
                if k in (
                    "author",
                    "date",
                    "amend",
                    "signoff",
                    "cleanup",
                    "reset_author",
                    "gpg_sign",
                    "no_gpg_sign",
                    "reedit_message",
                    "reuse_message",
                    "fixup",
                    "quiet",
                ):
                    if v is None:
                        continue
                    if isinstance(v, bool):
                        if v is True:
                            args.append("--%s" % k.replace("_", "-"))
                    else:
                        args.append("--{}={}".format(k.replace("_", "-"), v))

            to_add = [f for f in files if os.path.exists(f)]
            if to_add:
                system(
                    ['git', 'add', '-f', '-N', '--'] + to_add,
                    onerr=Abort,
                    errprefix=_("add failed"),
                )
            if not opts["message"]:
                system(
                    ['git', 'commit'] + args + ['--'] + list(files),
                    onerr=Abort,
                    errprefix=_("commit failed"),
                )
            else:
                system(
                    ['git', 'commit', '-F', msgfile.name] + args + ['--'] + list(files),
                    onerr=Abort,
                    errprefix=_("commit failed"),
                )
            # refresh the index so that gitk doesn’t show empty staged diffs
            systemcall(
                [
                    'git',
                    'update-index',
                    '--ignore-submodules',
                    '-q',
                    '--ignore-missing',
                    '--unmerged',
                    '--refresh',
                ],
            )

        finally:
            msgfile.unlink(missing_ok=True)


def main():
    prog = os.path.basename(sys.argv[0]).replace('-', ' ')
    version = importlib.metadata.version("git-crecord")

    subcommand = prog.split(' ')[-1].replace('.py', '')

    if subcommand == 'crecord':
        action = 'commit or stage'
    elif subcommand == 'cstage':
        action = 'stage'
    elif subcommand == 'cunstage':
        action = 'keep staged'

    parser = argparse.ArgumentParser(description='interactively select changes to %s' % action, prog=prog)
    parser.add_argument('--author', default=None, help='override author for commit')
    parser.add_argument('--date', default=None, help='override date for commit')
    parser.add_argument('-m', '--message', default=None, help='commit message')
    parser.add_argument(
        '-c', '--reedit-message', metavar='COMMIT', default=None,
        help='reuse and edit message from specified commit',
    )
    parser.add_argument(
        '-C', '--reuse-message', metavar='COMMIT', default=None,
        help='reuse message from specified commit',
    )
    parser.add_argument(
        '--fixup', metavar='COMMIT', default=None,
        help='create autosquash commit message to fixup specified commit',
    )
    parser.add_argument(
        '--reset-author', action='store_true', default=False,
        help='the commit is authored by me now (used with -C/-c/--amend)',
    )
    parser.add_argument('-s', '--signoff', action='store_true', default=False, help='add Signed-off-by:')
    parser.add_argument('--amend', action='store_true', default=False, help='amend previous commit')
    parser.add_argument(
        '-S', '--gpg-sign', metavar='KEY-ID', nargs='?', const=True, default=None,
        help='GPG sign commit',
    )
    parser.add_argument('--no-gpg-sign', action='store_true', default=False, help=argparse.SUPPRESS)
    parser.add_argument('-v', '--verbose', default=0, action='count', help='be more verbose')
    parser.add_argument('--debug', action='store_const', const=2, dest='verbose', help='be debuggingly verbose')
    parser.add_argument('--cleanup', default=None, help=argparse.SUPPRESS)
    parser.add_argument('--quiet', default=False, action='store_true', help='pass --quiet to git commit')
    parser.add_argument(
        '--confirm', default=False, action='store_true',
        help='show confirmation prompt after selecting changes',
    )
    parser.add_argument('--version', action='version', version=f'%(prog)s {version}')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--cached', '--staged', action='store_true', default=False, help=argparse.SUPPRESS)
    group.add_argument('--index', action='store_true', default=False, help=argparse.SUPPRESS)
    args = parser.parse_args()

    opts = vars(args)
    opts['operation'] = subcommand

    if opts['fixup'] and opts['fixup'].startswith('reword:'):
        parser.error(_("Creating reword-only commits is not supported."))

    if subcommand == 'cstage':
        opts['index'] = True

    if subcommand == 'cunstage':
        opts['cached'] = True

    repo = GitRepo(".")
    ui = Ui(repo)
    ui.setdebuglevel(opts['verbose'])

    os.chdir(repo.path)

    try:
        crecord_core.dorecord(ui, repo, **opts)
    except Abort as inst:
        sinst = str(inst)
        if opts['quiet'] and 'commit failed' in sinst:
            sys.exit(5)
        else:
            sys.stderr.write(_("abort: %s\n") % sinst)
            sys.exit(1)



# Chunk selector text user interface
#
# Copyright 2008—2011, 2014 Mark Edgington <edgimar@gmail.com>
# Copyright 2016, 2018—2022 Andrej Shadura <andrew@shadura.me>
#
# This software may be used and distributed according to the terms of
# the GNU General Public License, incorporated herein by reference.
#
# Much of this extension is based on Bryan O'Sullivan's record extension.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from __future__ import annotations

import curses
import fcntl
import re
import signal
import struct
import sys
import termios
from collections.abc import MutableSequence, Sequence
from gettext import gettext as _
from textwrap import dedent

from . import encoding, util
from .crpatch import Header, Hunk, HunkLine, PatchRoot

_origstdout = sys.__stdout__  # used by gethw()


def gethw() -> tuple[int, int]:
    """
    Magically get the current height and width of the window (without initscr)

    This is a rip-off of a rip-off - taken from the bpython code.  It is
    useful / necessary because otherwise curses.initscr() must be called,
    which can leave the terminal in a nasty state after exiting.

    """
    h, w = struct.unpack(
        "hhhh", fcntl.ioctl(_origstdout, termios.TIOCGWINSZ, b"\0" * 8),
    )[0:2]
    return h, w


def chunkselector(opts, headerlist, ui):
    """
    Curses interface to get selection of chunks, and mark the applied flags
    of the chosen chunks.

    """
    chunkselector = CursesChunkSelector(headerlist, ui)
    # This is required for ncurses to display non-ASCII characters in default user
    # locale encoding correctly.  --immerrr
    import locale

    locale.setlocale(locale.LC_ALL, '')

    class DummyStdscr:
        def clear(self):
            pass

        def refresh(self):
            pass

    chunkselector.stdscr = DummyStdscr()

    f = signal.getsignal(signal.SIGTSTP)
    curses.wrapper(chunkselector.main, opts)
    if chunkselector.initerr is not None:
        raise util.Abort(chunkselector.initerr)
    # ncurses does not restore signal handler for SIGTSTP
    signal.signal(signal.SIGTSTP, f)


header_messages = {
    # {operation: text}
    'crecord': _('Select hunks to commit'),
    'cstage': _('Select hunks to stage'),
    'cunstage': _('Select hunks to keep'),
}

main_operation_messages = {
    # {operation: text}
    'crecord': _('c: commit'),
    'cstage': _('s: stage'),
    'cunstage': None,  # TODO: not implemented!
}

confirm_messages = {
    'crecord': _('Are you sure you want to commit the selected changes [Yn]?'),
    'cstage': _('Are you sure you want to stage the selected changes [Yn]?'),
    'cunstage': _('Are you sure you want to unstage the unselected changes [Yn]?'),
}


class CursesChunkSelector:
    def __init__(self, headerlist, ui):
        # put the headers into a patch object
        self.headerlist = PatchRoot(headerlist)

        self.ui = ui

        self.errorstr = None
        # list of all chunks
        self.chunklist = []
        for h in headerlist:
            self.chunklist.append(h)
            self.chunklist.extend(h.hunks)

        # dictionary mapping (fgcolor, bgcolor) pairs to the
        # corresponding curses color-pair value.
        self.colorpairs = {}
        # maps custom nicknames of color-pairs to curses color-pair values
        self.colorpairnames = {}

        self.usecolor = True
        # the currently selected header, hunk, or hunk-line
        self.currentselecteditem = self.headerlist[0]

        # updated when printing out patch-display -- the 'lines' here are the
        # line positions *in the pad*, not on the screen.
        self.selecteditemstartline = 0
        self.selecteditemendline = None

        # define indentation levels
        self.headerindentnumchars = 0
        self.hunkindentnumchars = 3
        self.hunklineindentnumchars = 6

        # the first line of the pad to print to the screen
        self.firstlineofpadtoprint = 0

        # keeps track of the number of lines in the pad
        self.numpadlines = None

        self.numstatuslines = 1

        # keep a running count of the number of lines printed to the pad
        # (used for determining when the selected item begins/ends)
        self.linesprintedtopadsofar = 0

        # the first line of the pad which is visible on the screen
        self.firstlineofpadtoprint = 0

        # if the last 'toggle all' command caused all changes to be applied
        self.waslasttoggleallapplied = True

    def handlefirstlineevent(self):
        """Handle 'g' to navigate to the top most file in the ncurses window."""
        self.currentselecteditem = self.headerlist[0]
        currentitem = self.currentselecteditem
        # select the parent item recursively until we're at a header
        while True:
            nextitem = currentitem.parentitem()
            if nextitem is None:
                break
            else:
                currentitem = nextitem

        self.currentselecteditem = currentitem

    def handlelastlineevent(self):
        """Handle 'G' to navigate to the bottom most file/hunk/line.

        The behaviour depends on the whether the fold is active or not.

        If the bottom most file is folded, it navigates to that file and stops there.
        If the bottom most file is unfolded, it navigates to the bottom most hunk in
        that file and stops there. If the bottom most hunk is unfolded, it navigates to
        the bottom most line in that hunk.
        """
        currentitem = self.currentselecteditem
        nextitem = currentitem.nextitem()
        # select the child item recursively until we're at a footer
        while nextitem is not None:
            nextitem = currentitem.nextitem()
            if nextitem is None:
                break
            else:
                currentitem = nextitem

        self.currentselecteditem = currentitem
        self.recenterdisplayedarea()

    def uparrowevent(self):
        """
        Try to select the previous item to the current item that has the
        most-indented level.  For example, if a hunk is selected, try to select
        the last HunkLine of the hunk prior to the selected hunk.  Or, if
        the first HunkLine of a hunk is currently selected, then select the
        hunk itself.
        """
        currentitem = self.currentselecteditem

        nextitem = currentitem.previtem()

        if nextitem is None:
            # if no parent item (i.e. currentitem is the first header), then
            # no change...
            nextitem = currentitem

        self.currentselecteditem = nextitem
        self.recenterdisplayedarea()

    def uparrowshiftevent(self):
        """
        Select (if possible) the previous item on the same level as the
        currently selected item.  Otherwise, select (if possible) the
        parent-item of the currently selected item.
        """
        currentitem = self.currentselecteditem
        nextitem = currentitem.prevsibling()
        # if there's no previous sibling, try choosing the parent
        if nextitem is None:
            nextitem = currentitem.parentitem()
        if nextitem is None:
            # if no parent item (i.e. currentitem is the first header), then
            # no change...
            nextitem = currentitem

        self.currentselecteditem = nextitem
        self.recenterdisplayedarea()

    def downarrowevent(self):
        """
        Try to select the next item to the current item that has the
        most-indented level.  For example, if a hunk is selected, select
        the first HunkLine of the selected hunk.  Or, if the last HunkLine of
        a hunk is currently selected, then select the next hunk, if one exists,
        or if not, the next header if one exists.
        """
        currentitem = self.currentselecteditem

        nextitem = currentitem.nextitem()
        # if there's no next item, keep the selection as-is
        if nextitem is None:
            nextitem = currentitem

        self.currentselecteditem = nextitem
        self.recenterdisplayedarea()

    def downarrowshiftevent(self):
        """
        Select (if possible) the next item on the same level as the currently
        selected item.  Otherwise, select (if possible) the next item on the
        same level as the parent item of the currently selected item.
        """
        currentitem = self.currentselecteditem
        nextitem = currentitem.nextsibling()
        # if there's no next sibling, try choosing the parent's nextsibling
        if nextitem is None:
            try:
                nextitem = currentitem.parentitem().nextsibling()
            except AttributeError:
                # parentitem returned None, so nextsibling() can't be called
                nextitem = None
        if nextitem is None:
            # if parent has no next sibling, then no change...
            nextitem = currentitem

        self.currentselecteditem = nextitem
        self.recenterdisplayedarea()

    def rightarrowevent(self):
        """Select (if possible) the first of this item's child-items."""
        currentitem = self.currentselecteditem
        nextitem = currentitem.firstchild()

        # turn off folding if we want to show a child-item
        if currentitem.folded:
            self.togglefolded(currentitem)

        if nextitem is None:
            # if no next item on parent-level, then no change...
            nextitem = currentitem

        self.currentselecteditem = nextitem
        self.recenterdisplayedarea()

    def leftarrowevent(self):
        """Select parent or fold.

        If the current item can be folded (i.e. it is an unfolded header or
        hunk), then fold it. Otherwise, try to select (if possible) the parent
        of this item.
        """
        currentitem = self.currentselecteditem

        # try to fold the item
        if not isinstance(currentitem, HunkLine):
            if not currentitem.folded:
                self.togglefolded(item=currentitem)
                return

        # if it can't be folded, try to select the parent item
        nextitem = currentitem.parentitem()

        if nextitem is None:
            # if no item on parent-level, then no change...
            nextitem = currentitem
            if not nextitem.folded:
                self.togglefolded(item=nextitem)

        self.currentselecteditem = nextitem
        self.recenterdisplayedarea()

    def leftarrowshiftevent(self):
        """
        Select the header of the current item (or fold current item if the
        current item is already a header).

        """
        currentitem = self.currentselecteditem

        if isinstance(currentitem, Header):
            if not currentitem.folded:
                self.togglefolded(item=currentitem)
                return

        # select the parent item recursively until we're at a header
        while True:
            nextitem = currentitem.parentitem()
            if nextitem is None:
                break
            else:
                currentitem = nextitem

        self.currentselecteditem = currentitem
        self.recenterdisplayedarea()

    def updatescroll(self):
        """Scroll the screen to fully show the currently-selected"""
        selstart = self.selecteditemstartline
        selend = self.selecteditemendline
        padstart = self.firstlineofpadtoprint
        padend = padstart + self.yscreensize - self.numstatuslines - 1
        # 'buffered' pad start/end values which scroll with a certain
        # top/bottom context margin
        padstartbuffered = padstart + 3
        padendbuffered = padend - 3

        if selend > padendbuffered:
            self.scrolllines(selend - padendbuffered)
        elif selstart < padstartbuffered:
            # negative values scroll in pgup direction
            self.scrolllines(selstart - padstartbuffered)

    def scrolllines(self, numlines):
        """Scroll the screen up (down) by numlines when numlines >0 (<0)."""
        self.firstlineofpadtoprint += numlines
        if self.firstlineofpadtoprint < 0:
            self.firstlineofpadtoprint = 0
        if self.firstlineofpadtoprint > self.numpadlines - 1:
            self.firstlineofpadtoprint = self.numpadlines - 1

    def toggleapply(self, item=None):
        """
        Toggle the applied flag of the specified item.  If no item is specified,
        toggle the flag of the currently selected item.

        """
        if item is None:
            item = self.currentselecteditem

        item.applied = not item.applied

        if isinstance(item, Header):
            item.partial = False
            if item.applied:
                # apply all its hunks
                for hnk in item.hunks:
                    hnk.applied = True
                    # apply all their hunklines
                    for hunkline in hnk.changedlines:
                        hunkline.applied = True
            else:
                # un-apply all its hunks
                for hnk in item.hunks:
                    hnk.applied = False
                    hnk.partial = False
                    # un-apply all their hunklines
                    for hunkline in hnk.changedlines:
                        hunkline.applied = False
        elif isinstance(item, Hunk):
            item.partial = False
            # apply all it's hunklines
            for hunkline in item.changedlines:
                hunkline.applied = item.applied

            siblingappliedstatus = [hnk.applied for hnk in item.header.hunks]
            allsiblingsapplied = all(siblingappliedstatus)
            nosiblingsapplied = not any(siblingappliedstatus)

            siblingspartialstatus = [hnk.partial for hnk in item.header.hunks]
            somesiblingspartial = any(siblingspartialstatus)

            # cases where applied or partial should be removed from header

            # if no 'sibling' hunks are applied (including this hunk)
            if nosiblingsapplied:
                if not item.header.special():
                    item.header.applied = False
                    item.header.partial = False
            else:  # some/all parent siblings are applied
                item.header.applied = True
                item.header.partial = somesiblingspartial or not allsiblingsapplied
        elif isinstance(item, HunkLine):
            siblingappliedstatus = [ln.applied for ln in item.hunk.changedlines]
            allsiblingsapplied = all(siblingappliedstatus)
            nosiblingsapplied = not any(siblingappliedstatus)

            # if no 'sibling' lines are applied
            if nosiblingsapplied:
                item.hunk.applied = False
                item.hunk.partial = False
            elif allsiblingsapplied:
                item.hunk.applied = True
                item.hunk.partial = False
            else:  # some siblings applied
                item.hunk.applied = True
                item.hunk.partial = True

            parentsiblingsapplied = [hnk.applied for hnk in item.hunk.header.hunks]
            noparentsiblingsapplied = not any(parentsiblingsapplied)
            allparentsiblingsapplied = all(parentsiblingsapplied)

            parentsiblingspartial = [hnk.partial for hnk in item.hunk.header.hunks]
            someparentsiblingspartial = any(parentsiblingspartial)

            # if all parent hunks are not applied, un-apply header
            if noparentsiblingsapplied:
                if not item.hunk.header.special():
                    item.hunk.header.applied = False
                    item.hunk.header.partial = False
            # set the applied and partial status of the header if needed
            else:  # some/all parent siblings are applied
                item.hunk.header.applied = True
                item.hunk.header.partial = someparentsiblingspartial or not allparentsiblingsapplied

    def toggleall(self):
        """Toggle the applied flag of all items."""
        if self.waslasttoggleallapplied:  # then unapply them this time
            for item in self.headerlist:
                if item.applied:
                    self.toggleapply(item)
        else:
            for item in self.headerlist:
                if not item.applied:
                    self.toggleapply(item)
        self.waslasttoggleallapplied = not self.waslasttoggleallapplied

    def togglefolded(self, item=None, foldparent=False):
        """Toggle folded flag of specified item (defaults to currently selected)"""
        if item is None:
            item = self.currentselecteditem
        if foldparent or (isinstance(item, Header) and item.neverunfolded):
            if not isinstance(item, Header):
                # we need to select the parent item in this case
                self.currentselecteditem = item = item.parentitem()
            elif item.neverunfolded:
                item.neverunfolded = False

            # also fold any foldable children of the parent/current item
            if isinstance(item, Header):  # the original OR 'new' item
                for child in item.allchildren():
                    child.folded = not item.folded

        if isinstance(item, (Header, Hunk)):
            item.folded = not item.folded

    def alignstring(self, instr, window):
        """
        Add whitespace to the end of a string in order to make it fill
        the screen in the x direction.  The current cursor position is
        taken into account when making this calculation.  The string can span
        multiple lines.

        """
        y, xstart = window.getyx()
        width = self.xscreensize
        # turn tabs into spaces
        instr = instr.expandtabs(4)
        strwidth = encoding.ucolwidth(instr)
        numspaces = width - ((strwidth + xstart) % width)
        return instr + " " * numspaces

    def printstring(
        self,
        window,
        text,
        fgcolor=None,
        bgcolor=None,
        pair=None,
        pairname=None,
        attrlist=None,
        towin=True,
        align=True,
        showwhtspc=False,
    ):
        """
        Print the string, text, with the specified colors and attributes, to
        the specified curses window object.

        The foreground and background colors are of the form
        curses.COLOR_XXXX, where XXXX is one of: [BLACK, BLUE, CYAN, GREEN,
        MAGENTA, RED, WHITE, YELLOW].  If pairname is provided, a color
        pair will be looked up in the self.colorpairnames dictionary.

        attrlist is a list containing text attributes in the form of
        curses.A_XXXX, where XXXX can be: [BOLD, DIM, NORMAL, STANDOUT,
        UNDERLINE].

        If align == True, whitespace is added to the printed string such that
        the string stretches to the right border of the window.

        If showwhtspc == True, trailing whitespace of a string is highlighted.

        """
        # preprocess the text, converting tabs to spaces
        text = text.expandtabs(4)
        # Strip \n, and convert control characters to ^[char] representation
        text = re.sub(
            r'[\x00-\x08\x0a-\x1f]',
            lambda m: '^' + chr(ord(m.group()) + 64), text.strip('\n'),
        )

        if pair is not None:
            colorpair = pair
        elif pairname is not None:
            colorpair = self.colorpairnames[pairname]
        else:
            if fgcolor is None:
                fgcolor = -1
            if bgcolor is None:
                bgcolor = -1
            if (fgcolor, bgcolor) in self.colorpairs:
                colorpair = self.colorpairs[(fgcolor, bgcolor)]
            else:
                colorpair = self.getcolorpair(fgcolor, bgcolor)
        # add attributes if possible
        if attrlist is None:
            attrlist = []
        if colorpair < 256:
            # then it is safe to apply all attributes
            for textattr in attrlist:
                colorpair |= textattr
        else:
            # just apply a select few (safe?) attributes
            for textattr in (curses.A_UNDERLINE, curses.A_BOLD):
                if textattr in attrlist:
                    colorpair |= textattr

        y, xstart = self.chunkpad.getyx()
        t = ""  # variable for counting lines printed
        # if requested, show trailing whitespace
        if showwhtspc:
            origlen = len(text)
            text = text.rstrip(' \n')  # tabs have already been expanded
            strippedlen = len(text)
            numtrailingspaces = origlen - strippedlen

        if towin:
            window.addstr(text, colorpair)
        t += text

        if showwhtspc:
            wscolorpair = colorpair | curses.A_REVERSE
            if towin:
                for i in range(numtrailingspaces):
                    window.addch(curses.ACS_CKBOARD, wscolorpair)
            t += " " * numtrailingspaces

        if align:
            if towin:
                extrawhitespace = self.alignstring("", window)
                window.addstr(extrawhitespace, colorpair)
            else:
                # need to use t, since the x position hasn't incremented
                extrawhitespace = self.alignstring(t, window)
            t += extrawhitespace

        # is reset to 0 at the beginning of printitem()

        linesprinted = (xstart + len(t)) // self.xscreensize
        self.linesprintedtopadsofar += linesprinted
        return t

    def _getstatuslinesegments(self):
        """-> [str]. return segments"""
        selected = self.currentselecteditem.applied
        segments = [
            header_messages[self.opts['operation']],
            '-',
            _('[x]=selected **=collapsed'),
            main_operation_messages[self.opts['operation']],
            _('q: abort'),
            _('arrow keys: move/expand/collapse'),
            _('space: deselect') if selected else _('space: select'),
            _('?: help'),
        ]
        return segments

    def _getstatuslines(self) -> Sequence[str]:
        """() -> [str]. return short help used in the top status window"""
        if self.errorstr is not None:
            lines = [self.errorstr, _('Press any key to continue')]
        else:
            # wrap segments to lines
            segments = self._getstatuslinesegments()
            width = self.xscreensize
            lines = []
            lastwidth = width
            for s in segments:
                w = encoding.ucolwidth(s)
                sep = ' ' * (1 + (s and s[0] not in '-['))
                if lastwidth + w + len(sep) >= width:
                    lines.append(s)
                    lastwidth = w
                else:
                    lines[-1] += sep + s
                    lastwidth += w + len(sep)
        if len(lines) != self.numstatuslines:
            self.numstatuslines = len(lines)
            self.statuswin.resize(self.numstatuslines, self.xscreensize)
        return [util.ellipsis(line, self.xscreensize - 1) for line in lines]

    def updatescreen(self):
        self.statuswin.erase()
        self.chunkpad.erase()

        printstring = self.printstring

        # print out the status lines at the top
        try:
            for line in self._getstatuslines():
                printstring(self.statuswin, line, pairname="legend")
            self.statuswin.refresh()
        except curses.error:
            pass
        if self.errorstr is not None:
            return

        # print out the patch in the remaining part of the window
        try:
            self.printitem()
            self.updatescroll()
            self.chunkpad.refresh(
                self.firstlineofpadtoprint, 0,
                self.numstatuslines, 0,
                self.yscreensize - self.numstatuslines,
                self.xscreensize,
            )
        except curses.error:
            pass

    def getstatusprefixstring(self, item):
        """
        Create a string to prefix a line with which indicates whether 'item'
        is applied and/or folded.

        """
        # create checkbox string
        if item.applied:
            if not isinstance(item, HunkLine) and item.partial:
                checkbox = "[~]"
            else:
                checkbox = "[x]"
        else:
            checkbox = "[ ]"

        try:
            if item.folded:
                checkbox += "**"
                if isinstance(item, Header):
                    # one of "M", "A", or "D" (modified, added, deleted)
                    filestatus = item.changetype

                    checkbox += filestatus + " "
            else:
                checkbox += "  "
                if isinstance(item, Header):
                    # add two more spaces for headers
                    checkbox += "  "
        except AttributeError:  # not foldable
            checkbox += "  "

        return checkbox

    def printheader(
        self, header: Header, selected=False, towin=True, ignorefolding=False,
    ):
        """
        Print the header to the pad.  If countLines is True, don't print
        anything, but just count the number of lines which would be printed.

        """
        outstr = ""
        text = header.prettystr()
        chunkindex = self.chunklist.index(header)

        if chunkindex != 0 and not header.folded:
            # add separating line before headers
            outstr += self.printstring(
                self.chunkpad, "_" * self.xscreensize, towin=towin, align=False,
            )
        # select color-pair based on if the header is selected
        colorpair = self.getcolorpair(
            name=selected and "selected" or "normal", attrlist=[curses.A_BOLD],
        )

        # print out each line of the chunk, expanding it to screen width

        # number of characters to indent lines on this level by
        indentnumchars = 0
        checkbox = self.getstatusprefixstring(header)
        if not header.folded or ignorefolding:
            textlist = text.split("\n")
            linestr = checkbox + textlist[0]
        else:
            linestr = checkbox + header.filename()
        outstr += self.printstring(self.chunkpad, linestr, pair=colorpair, towin=towin)
        if not header.folded or ignorefolding:
            if len(textlist) > 1:
                for line in textlist[1:]:
                    linestr = " " * (indentnumchars + len(checkbox)) + line
                    outstr += self.printstring(
                        self.chunkpad, linestr, pair=colorpair, towin=towin,
                    )

        return outstr

    def printhunklinesbefore(
        self, hunk: Hunk, selected=False, towin=True, ignorefolding=False,
    ):
        """Print lines including the start/end line indicator."""
        outstr = ""
        # where hunk is in list of siblings
        hunkindex = hunk.header.hunks.index(hunk)

        if hunkindex != 0:
            # add separating line before headers
            outstr += self.printstring(
                self.chunkpad, " " * self.xscreensize, towin=towin, align=False,
            )

        colorpair = self.getcolorpair(
            name=selected and "selected" or "normal", attrlist=[curses.A_BOLD],
        )

        # print out from-to line with checkbox
        checkbox = self.getstatusprefixstring(hunk)

        lineprefix = " " * self.hunkindentnumchars + checkbox
        frtoline = "   " + hunk.getfromtoline().decode("UTF-8", errors="hexreplace").strip("\n")

        outstr += self.printstring(
            self.chunkpad, lineprefix, towin=towin, align=False,
        )  # add uncolored checkbox/indent
        outstr += self.printstring(self.chunkpad, frtoline, pair=colorpair, towin=towin)

        if hunk.folded and not ignorefolding:
            # skip remainder of output
            return outstr

        # print out lines of the chunk preceding changed-lines
        for line in hunk.before:
            linestr = " " * (self.hunklineindentnumchars + len(checkbox)) + line.decode("UTF-8", errors="hexreplace")
            outstr += self.printstring(self.chunkpad, linestr, towin=towin)

        return outstr

    def printhunklinesafter(self, hunk: Hunk, towin=True, ignorefolding=False):
        outstr = ""
        if hunk.folded and not ignorefolding:
            return outstr

        # a bit superfluous, but to avoid hard-coding indent amount
        checkbox = self.getstatusprefixstring(hunk)
        for line in hunk.after:
            linestr = " " * (self.hunklineindentnumchars + len(checkbox)) + line.decode("UTF-8", errors="hexreplace")
            outstr += self.printstring(self.chunkpad, linestr, towin=towin)

        return outstr

    def printhunkchangedline(self, hunkline: HunkLine, selected=False, towin=True):
        outstr = ""
        checkbox = self.getstatusprefixstring(hunkline)

        linestr = hunkline.prettystr().strip("\n")

        # select color-pair based on whether line is an addition/removal
        if selected:
            colorpair = self.getcolorpair(name="selected")
        elif linestr.startswith("+"):
            colorpair = self.getcolorpair(name="addition")
        elif linestr.startswith("-"):
            colorpair = self.getcolorpair(name="deletion")
        elif linestr.startswith("\\"):
            colorpair = self.getcolorpair(name="normal")

        lineprefix = " " * self.hunklineindentnumchars + checkbox
        outstr += self.printstring(
            self.chunkpad, lineprefix, towin=towin, align=False,
        )  # add uncolored checkbox/indent
        outstr += self.printstring(
            self.chunkpad, linestr, pair=colorpair, towin=towin, showwhtspc=True,
        )
        return outstr

    def printitem(
        self, item=None, ignorefolding=False, recursechildren=True, towin=True,
    ):
        """Print the specified item.

        If item is not specified, then print the entire patch.
        (hiding folded elements, etc. -- see __printitem() docstring)
        """
        if item is None:
            item = self.headerlist
        if recursechildren:
            self.linesprintedtopadsofar = 0

        outstr = []
        self.__printitem(
            item, ignorefolding, recursechildren, outstr, towin=towin,
        )
        return ''.join(outstr)

    def outofdisplayedarea(self):
        y, _ = self.chunkpad.getyx()  # cursor location
        # * 2 here works but an optimization would be the max number of
        # consecutive non selectable lines
        # i.e the max number of context line for any hunk in the patch
        miny = min(0, self.firstlineofpadtoprint - self.yscreensize)
        maxy = self.firstlineofpadtoprint + self.yscreensize * 2
        return y < miny or y > maxy

    def handleselection(self, item, recursechildren):
        selected = item is self.currentselecteditem
        if selected and recursechildren:
            # assumes line numbering starting from line 0
            self.selecteditemstartline = self.linesprintedtopadsofar
            selecteditemlines = self.getnumlinesdisplayed(item, recursechildren=False)
            self.selecteditemendline = self.selecteditemstartline + selecteditemlines - 1

        return selected

    def __printitem(
        self,
        item: PatchRoot | Header | Hunk | HunkLine,
        ignorefolding: bool,
        recursechildren: bool,
        outstr: MutableSequence[str],
        towin: bool = True,
    ):
        """
        Recursive method for printing out patch/header/hunk/hunk-line data to
        screen.  Also returns a string with all of the content of the displayed
        patch (not including coloring, etc.).

        If ignorefolding is True, then folded items are printed out.

        If recursechildren is False, then only print the item without its
        child items.

        """
        if towin and self.outofdisplayedarea():
            return

        selected = self.handleselection(item, recursechildren)

        # Patch object is a list of headers
        if isinstance(item, PatchRoot):
            if recursechildren:
                for hdr in item:
                    self.__printitem(hdr, ignorefolding, recursechildren, outstr, towin)
        # TODO: eliminate all isinstance() calls
        if isinstance(item, Header):
            outstr.append(
                self.printheader(
                    item, selected, towin=towin, ignorefolding=ignorefolding,
                ),
            )
            if recursechildren:
                for hnk in item.hunks:
                    self.__printitem(hnk, ignorefolding, recursechildren, outstr, towin)
        elif isinstance(item, Hunk) and ((not item.header.folded) or ignorefolding):
            # print the hunk data which comes before the changed-lines
            outstr.append(
                self.printhunklinesbefore(
                    item, selected, towin=towin, ignorefolding=ignorefolding,
                ),
            )
            if recursechildren:
                for line in item.changedlines:
                    self.__printitem(
                        line, ignorefolding, recursechildren, outstr, towin,
                    )
                outstr.append(
                    self.printhunklinesafter(
                        item, towin=towin, ignorefolding=ignorefolding,
                    ),
                )
        elif isinstance(item, HunkLine) and ((not item.hunk.folded) or ignorefolding):
            outstr.append(self.printhunkchangedline(item, selected, towin=towin))

        return outstr

    def getnumlinesdisplayed(
        self, item=None, ignorefolding=False, recursechildren=True,
    ):
        """
        Return the number of lines which would be displayed if the item were
        to be printed to the display.  The item will NOT be printed to the
        display (pad).
        If no item is given, assume the entire patch.
        If ignorefolding is True, folded items will be unfolded when counting
        the number of lines.

        """
        # temporarily disable printing to windows by printstring
        patchdisplaystring = self.printitem(
            item, ignorefolding, recursechildren, towin=False,
        )
        numlines = len(patchdisplaystring) // self.xscreensize
        return numlines

    def sigwinchhandler(self, n, frame):
        """Handle window resizing"""
        try:
            curses.endwin()
            self.yscreensize, self.xscreensize = gethw()
            self.statuswin.resize(self.numstatuslines, self.xscreensize)
            self.numpadlines = self.getnumlinesdisplayed(ignorefolding=True) + 1
            self.chunkpad = curses.newpad(self.numpadlines, self.xscreensize)

        except curses.error:
            pass

    def getcolorpair(self, fgcolor=None, bgcolor=None, name=None, attrlist=None):
        """
        Get a curses color pair, adding it to self.colorPairs if it is not
        already defined.  An optional string, name, can be passed as a shortcut
        for referring to the color-pair.  By default, if no arguments are
        specified, the white foreground / black background color-pair is
        returned.

        It is expected that this function will be used exclusively for
        initializing color pairs, and NOT curses.init_pair().

        attrlist is used to 'flavor' the returned color-pair.  This information
        is not stored in self.colorpairs.  It contains attribute values like
        curses.A_BOLD.

        """
        if (name is not None) and name in self.colorpairnames:
            # then get the associated color pair and return it
            colorpair = self.colorpairnames[name]
        else:
            if fgcolor is None:
                fgcolor = -1
            if bgcolor is None:
                bgcolor = -1
            if (fgcolor, bgcolor) in self.colorpairs:
                colorpair = self.colorpairs[(fgcolor, bgcolor)]
            else:
                pairindex = len(self.colorpairs) + 1
                if self.usecolor:
                    curses.init_pair(pairindex, fgcolor, bgcolor)
                    colorpair = self.colorpairs[(fgcolor, bgcolor)] = curses.color_pair(
                        pairindex,
                    )
                    if name is not None:
                        self.colorpairnames[name] = curses.color_pair(pairindex)
                else:
                    cval = 0
                    if name is not None:
                        if name == 'selected':
                            cval = curses.A_REVERSE
                        self.colorpairnames[name] = cval
                    colorpair = self.colorpairs[(fgcolor, bgcolor)] = cval

        # add attributes if possible
        if attrlist is None:
            attrlist = []
        if colorpair < 256:
            # then it is safe to apply all attributes
            for textattr in attrlist:
                colorpair |= textattr
        else:
            # just apply a select few (safe?) attributes
            for textattrib in (curses.A_UNDERLINE, curses.A_BOLD):
                if textattrib in attrlist:
                    colorpair |= textattrib
        return colorpair

    def helpwindow(self):
        """Print a help window to the screen.  Exit after any keypress."""
        helptext = """            [press any key to return to the patch-display]

crecord allows you to interactively choose among the changes you have made,
and confirm only those changes you select for further processing by the command
you are running (commit/stage/unstage), after confirming the selected
changes, the unselected changes are still present in your working copy, so you
can use crecord multiple times to split large changes into smaller changesets.
The following are valid keystrokes:

                [SPACE] : (un-)select item ([~]/[X] = partly/fully applied)
                      A : (un-)select all items
    Up/Down-arrow [k/j] : go to previous/next unfolded item
        PgUp/PgDn [K/J] : go to previous/next item of same type
 Right/Left-arrow [l/h] : go to child item / parent item
 Shift-Left-arrow   [H] : go to parent header / fold selected header
                      g : go to the top
                      G : go to the bottom
                      f : fold / unfold item, hiding/revealing its children
                      F : fold / unfold parent item and all of its ancestors
                 ctrl-l : scroll the selected line to the top of the screen
                      a : toggle amend mode
                      c : commit selected changes
                      s : stage selected changes
                      r : review/edit and commit selected changes
                      q : quit without committing (no changes will be made)
                      ? : help (what you're currently reading)"""

        helpwin = curses.newwin(self.yscreensize, 0, 0, 0)
        helplines = helptext.split("\n")
        helplines = helplines + [" "] * (
            self.yscreensize - self.numstatuslines - len(helplines) - 1
        )
        try:
            self.printstring(helpwin, helplines[0], pairname="legend")
            for line in helplines[1:]:
                self.printstring(helpwin, line, pairname="normal")
        except curses.error:
            pass
        helpwin.refresh()
        try:
            helpwin.getkey()
        except curses.error:
            pass

    def confirmationwindow(self, windowtext):
        """Display an informational window, then wait for and return a keypress."""
        lines = windowtext.split("\n")
        confirmwin = curses.newwin(len(lines), 0, 0, 0)
        try:
            for line in lines:
                self.printstring(confirmwin, line, pairname="selected")
        except curses.error:
            pass
        try:
            response = chr(confirmwin.getch())
        except ValueError:
            response = None

        return response

    def confirmcommit(self, review=False):
        """Ask for 'Y' to be pressed to confirm selected. Return True if
        confirmed."""
        if not self.opts['confirm']:
            return True

        if review:
            confirmtext = dedent(
                """
                If you answer yes to the following, the your currently chosen patch chunks
                will be loaded into an editor.  You may modify the patch from the editor, and
                save the changes if you wish to change the patch.  Otherwise, you can just
                close the editor without saving to accept the current patch as-is.
                
                NOTE: don't add/remove lines unless you also modify the range information.
                      Failing to follow this rule will result in the commit aborting.
                
                Are you sure you want to review/edit and confirm the selected changes [Yn]?
                """,
            )
        else:
            confirmtext = confirm_messages[self.opts['operation']]

        response = self.confirmationwindow(confirmtext)

        if response is None or len(response) == 0 or response == "\n":
            response = "y"
        return response.lower().startswith("y")

    def recenterdisplayedarea(self):
        """
        Recenter the screen.

        Once we scrolled with PgUp/PgDown, we can be pointing outside the
        display zone. We print the patch with towin=False to compute the
        location of the selected item even though it is outside the displayed
        zone, and then update the scroll.
        """
        self.printitem(towin=False)
        self.updatescroll()

    def toggleamend(self):
        """Toggle the amend flag.

        When the amend flag is set, a commit will modify the most recent
        commit, instead of creating a new commit.  Otherwise, a
        new commit will be created (the normal commit behavior).
        """
        if not self.opts.get('amend'):
            self.opts['amend'] = True
            msg = (
                "Amend option is turned on -- committing the currently "
                "selected changes will not create a new commit, but "
                "instead update the most recent commit.\n\n"
                "Press any key to continue."
            )
        else:
            self.opts['amend'] = False
            msg = (
                "Amend option is turned off -- committing the currently "
                "selected changes will create a new commit.\n\n"
                "Press any key to continue."
            )

        self.confirmationwindow(msg)

    def handlekeypressed(self, keypressed):
        """
        Perform actions based on pressed keys.

        Return true to exit the main loop.
        """
        if keypressed in ["k", "KEY_UP"]:
            self.uparrowevent()
        elif keypressed in ["K", "KEY_PPAGE"]:
            self.uparrowshiftevent()
        elif keypressed in ["j", "KEY_DOWN"]:
            self.downarrowevent()
        elif keypressed in ["J", "KEY_NPAGE"]:
            self.downarrowshiftevent()
        elif keypressed in ["l", "KEY_RIGHT"]:
            self.rightarrowevent()
        elif keypressed in ["h", "KEY_LEFT"]:
            self.leftarrowevent()
        elif keypressed in ["H", "KEY_SLEFT"]:
            self.leftarrowshiftevent()
        elif keypressed in ["q"]:
            raise util.Abort(_('user quit'))
        elif keypressed in ['a']:
            self.toggleamend()
        elif keypressed in ["c"]:
            self.opts['confirm'] |= self.opts['operation'] != 'crecord'
            if self.confirmcommit():
                self.opts['operation'] = 'crecord'
                return True
        elif keypressed in ["s"]:
            self.opts['confirm'] |= self.opts['operation'] != 'cstage'
            if self.confirmcommit():
                self.opts['operation'] = 'cstage'
                return True
        elif keypressed in ["r"]:
            if self.confirmcommit(review=True):
                self.opts['commit'] = True
                self.opts['crecord_reviewpatch'] = True
                return True
        elif keypressed in [' ']:
            self.toggleapply()
        elif keypressed in ['A']:
            self.toggleall()
        elif keypressed in ["f"]:
            self.togglefolded()
        elif keypressed in ["F"]:
            self.togglefolded(foldparent=True)
        elif keypressed in ["?"]:
            self.helpwindow()
            self.stdscr.clear()
            self.stdscr.refresh()
        elif len(keypressed) == 1 and curses.unctrl(keypressed) in [b"^L"]:
            # scroll the current line to the top of the screen, and redraw
            # everything
            self.scrolllines(self.selecteditemstartline)
            self.stdscr.clear()
            self.stdscr.refresh()
        elif keypressed in ["g", "KEY_HOME"]:
            self.handlefirstlineevent()
        elif keypressed in ["G", "KEY_END"]:
            self.handlelastlineevent()
        return False

    def main(self, stdscr, opts):
        """Configure and restore signal handler, and run the record UI.

        Method to be wrapped by curses.wrapper() for selecting chunks.
        """
        self.opts = opts

        origsigwinch = sentinel = object()
        if util.safehasattr(signal, 'SIGWINCH'):
            origsigwinch = signal.signal(signal.SIGWINCH, self.sigwinchhandler)
        try:
            return self._main(stdscr)
        finally:
            if origsigwinch is not sentinel:
                signal.signal(signal.SIGWINCH, origsigwinch)

    def _main(self, stdscr):
        """Configure the display and run the event loop."""
        self.stdscr = stdscr
        # error during initialization, cannot be printed in the curses
        # interface, it should be printed by the calling code
        self.initerr = None
        self.yscreensize, self.xscreensize = self.stdscr.getmaxyx()

        curses.start_color()
        try:
            curses.use_default_colors()
        except curses.error:
            self.usecolor = False

        # In some situations we may have some cruft left on the "alternate
        # screen" from another program (or previous iterations of ourself), and
        # we won't clear it if the scroll region is small enough to comfortably
        # fit on the terminal.
        self.stdscr.clear()

        # don't display the cursor
        try:
            curses.curs_set(0)
        except curses.error:
            pass

        # available colors: black, blue, cyan, green, magenta, white, yellow
        # init_pair(color_id, foreground_color, background_color)
        self.getcolorpair(None, None, name="normal")
        self.getcolorpair(curses.COLOR_WHITE, curses.COLOR_MAGENTA, name="selected")
        self.getcolorpair(curses.COLOR_RED, None, name="deletion")
        self.getcolorpair(curses.COLOR_GREEN, None, name="addition")
        self.getcolorpair(curses.COLOR_WHITE, curses.COLOR_BLUE, name="legend")
        # newwin([height, width,] begin_y, begin_x)
        self.statuswin = curses.newwin(self.numstatuslines, 0, 0, 0)
        self.statuswin.keypad(True)  # interpret arrow-key, etc. ESC sequences

        # figure out how much space to allocate for the chunk-pad which is
        # used for displaying the patch

        # stupid hack to prevent getnumlinesdisplayed from failing
        self.chunkpad = curses.newpad(1, self.xscreensize)

        # add 1 so to account for last line text reaching end of line
        self.numpadlines = self.getnumlinesdisplayed(ignorefolding=True) + 1

        try:
            self.chunkpad = curses.newpad(self.numpadlines, self.xscreensize)
        except curses.error:
            self.initerr = _('this diff is too large to be displayed')
            return
        # initialize selecteditemendline (initial start-line is 0)
        self.selecteditemendline = self.getnumlinesdisplayed(
            self.currentselecteditem, recursechildren=False,
        )

        # option which enables/disables patch-review (in editor) step
        self.opts['crecord_reviewpatch'] = False

        while True:
            self.updatescreen()
            try:
                keypressed = self.statuswin.getkey()
                if self.errorstr is not None:
                    self.errorstr = None
                    continue
            except curses.error:
                keypressed = "FOOBAR"
            if self.handlekeypressed(keypressed):
                break



# crecord.py
#
# Copyright 2008 Mark Edgington <edgimar@gmail.com>
#
# This software may be used and distributed according to the terms of
# the GNU General Public License, incorporated herein by reference.
#
# Much of this extension is based on Bryan O'Sullivan's record extension.



# Utility functions
#
#  Copyright 2006, 2015 Matt Mackall <mpm@selenic.com>
#  Copyright 2007 Eric St-Jean <esj@wwd.ca>
#  Copyright 2009, 2011 Mads Kiilerich <mads@kiilerich.com>
#  Copyright 2015 Pierre-Yves David <pierre-yves.david@fb.com>
#  Copyright 2016, 2022 Andrej Shadura <andrew@shadura.me>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version
#
# Some of these utilities were originally taken from Mercurial.
#
# SPDX-License-Identifier: GPL-2.0-or-later
from __future__ import annotations

import os
import shutil
import subprocess
import sys
from collections.abc import Sequence
from gettext import gettext as _
from pathlib import Path
from typing import Optional, overload

from .encoding import ucolwidth

closefds = os.name == 'posix'


def explainexit(code):
    """Return a 2-tuple (desc, code) describing a subprocess status.

    Codes from kill are negative - not os.system/wait encoding.
    """
    if (code < 0) and (os.name == 'posix'):
        return _("killed by signal %d") % -code, -code
    else:
        return _("exited with status %d") % code, code


class Abort(Exception):
    pass


def system(cmd, cwd=None, onerr=None, errprefix=None):
    try:
        sys.stdout.flush()
    except Exception:
        pass

    if isinstance(cmd, list):
        shell = False
        prog = os.path.basename(cmd[0])
    else:
        shell = True
        prog = os.path.basename(cmd.split(None, 1)[0])

    rc = subprocess.call(cmd, shell=shell, close_fds=closefds, cwd=cwd)
    if rc and onerr:
        errmsg = f"{prog} {explainexit(rc)[0]}"
        if errprefix:
            errmsg = f"{errprefix}: {errmsg}"
        raise onerr(errmsg)
    return rc


@overload
def systemcall(
    cmd: Sequence[str] | Sequence[bytes],
    encoding: str,
    dir: os.PathLike | str | None = None,
    onerr=None,
    errprefix=None,
) -> str:
    ...


@overload
def systemcall(
    cmd: Sequence[str] | Sequence[bytes],
    dir: os.PathLike | str | None = None,
    onerr=None,
    errprefix=None,
) -> bytes:
    ...


def systemcall(cmd, encoding=None, dir=None, onerr=None, errprefix=None):
    try:
        sys.stdout.flush()
    except Exception:  # noqa: B902
        pass

    p = subprocess.Popen(cmd, cwd=dir, stdout=subprocess.PIPE, close_fds=closefds)
    out = b''
    for line in iter(p.stdout.readline, b''):
        out = out + line
    p.wait()
    rc = p.returncode

    if rc and onerr:
        errmsg = f'{os.path.basename(cmd[0])} {explainexit(rc)[0]}'
        if errprefix:
            errmsg = f'{errprefix}: {errmsg}'
        raise onerr(errmsg)

    if encoding == "fs":
        return os.fsdecode(out)
    elif encoding:
        return out.decode(encoding)
    else:
        return out


def copyfile(src: str | Path, dest: str | Path, copystat=True):
    """Copy a file, preserving mode and optionally other stat info like atime/mtime"""
    if os.path.lexists(dest):
        os.unlink(dest)
    if os.path.islink(src):
        os.symlink(os.readlink(src), dest)
        # copytime is ignored for symlinks, but in general copytime isn't needed
        # for them anyway
    else:
        try:
            shutil.copyfile(src, dest)  # type: ignore
            if copystat:
                # copystat also copies mode
                shutil.copystat(src, dest)
            else:
                shutil.copymode(src, dest)
        except shutil.Error as inst:
            raise Abort(str(inst))


def ellipsis(text, maxlength=400):
    """Trim string to at most maxlength (default: 400) columns in display."""
    return trim(text, maxlength, ellipsis='...')


def trim(s, width, ellipsis='', leftside=False):
    """Trim string 's' to at most 'width' columns (including 'ellipsis').

    If 'leftside' is True, left side of string 's' is trimmed.
    'ellipsis' is always placed at trimmed side.

    >>> pluses = '+++'
    >>> encoding = 'utf-8'
    >>> t = '1234567890'
    >>> print(trim(t, 12, ellipsis=pluses))
    1234567890
    >>> print(trim(t, 10, ellipsis=pluses))
    1234567890
    >>> print(trim(t, 8, ellipsis=pluses))
    12345+++
    >>> print(trim(t, 8, ellipsis=pluses, leftside=True))
    +++67890
    >>> print(trim(t, 8))
    12345678
    >>> print(trim(t, 8, leftside=True))
    34567890
    >>> print(trim(t, 3, ellipsis=pluses))
    +++
    >>> print(trim(t, 1, ellipsis=pluses))
    +
    >>> t = '\u3042\u3044\u3046\u3048\u304a' # 2 x 5 = 10 columns
    >>> print(trim(t, 12, ellipsis=pluses))
    \u3042\u3044\u3046\u3048\u304a
    >>> print(trim(t, 10, ellipsis=pluses))
    \u3042\u3044\u3046\u3048\u304a
    >>> print(trim(t, 8, ellipsis=pluses))
    \u3042\u3044+++
    >>> print(trim(t, 8, ellipsis=pluses, leftside=True))
    +++\u3048\u304a
    >>> print(trim(t, 5))
    \u3042\u3044
    >>> print(trim(t, 5, leftside=True))
    \u3048\u304a
    >>> print(trim(t, 4, ellipsis=pluses))
    +++
    >>> print(trim(t, 4, ellipsis=pluses, leftside=True))
    +++
    """
    if ucolwidth(s) <= width:  # trimming is not needed
        return s

    width -= len(ellipsis)
    if width <= 0:  # no enough room even for ellipsis
        return ellipsis[:width + len(ellipsis)]

    # TODO: find a way to get rid of the lambdas
    if leftside:
        uslice = lambda i: s[i:]  # noqa: E731
        concat = lambda s: ellipsis + s  # noqa: E731
    else:
        uslice = lambda i: s[:-i]  # noqa: E731
        concat = lambda s: s + ellipsis  # noqa: E731
    for i in range(1, len(s)):
        usub = uslice(i)
        if ucolwidth(usub) <= width:
            return concat(usub)
    return ellipsis  # no enough room for multi-column characters


_notset = object()


def safehasattr(thing, attr):
    return getattr(thing, attr, _notset) is not _notset


def unescape_filename(filename: bytes) -> bytes:
    r"""Unescape a filename after Git mangled it for "git diff --git" line.

    >>> unescape_filename(b'a/\\321\\216\\321\\217')
    b'a/\xd1\x8e\xd1\x8f'
    >>> unescape_filename(b'a/\\\\')
    b'a/\\'
    >>> unescape_filename(b'a/file\\55name')
    b'a/file-name'
    """
    unescaped_unicode = filename.decode('unicode_escape')
    return bytes(ord(x) for x in unescaped_unicode)


def unwrap_filename(filename: bytes) -> bytes:
    r"""Unwrap a filename mangled by Git

    If the filename is in double quotes, remove them and unescape enclosed characters.
    Otherwise, return the input as is.

    >>> def apply(f, s: str) -> str:
    ...     return f(s.encode("UTF-8")).decode("UTF-8")
    >>> apply(unwrap_filename, 'a/filename')
    'a/filename'
    >>> apply(unwrap_filename, 'a/имя-файла')
    'a/имя-файла'
    >>> apply(unwrap_filename, '"a/file\\55name"')
    'a/file-name'
    >>> apply(unwrap_filename, '"a/им\\321\\217\55\\\\name"')
    'a/имя-\\name'
    """
    if filename.startswith(b'"') and filename.endswith(b'"'):
        return unescape_filename(filename[1:-1])
    else:
        return filename
